@@ -1,5 +1,6 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+import hashlib
 from decimal import Decimal
 
 from sql import Null
@@ -8,21 +9,19 @@ from sql.conditionals import Case, Coalesce
 from sql.functions import Abs
 
 from trytond import backend
+from trytond.i18n import gettext
 from trytond.pool import Pool, PoolMeta
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.pyson import Eval, If, Bool
+from trytond.pyson import Eval, If, Bool, Id
 from trytond.wizard import (Wizard, StateView, StateAction, StateTransition,
     Button)
 from trytond.transaction import Transaction
 from trytond.tools.multivalue import migrate_property
 from trytond.modules.company.model import CompanyValueMixin
 
+from .exceptions import BlockedWarning, GroupWarning
 from .payment import KINDS
 
-__all__ = ['MoveLine', 'PayLine', 'PayLineAskJournal',
-    'Configuration', 'ConfigurationPaymentGroupSequence',
-    'Invoice']
-
 
 class MoveLine(metaclass=PoolMeta):
     __name__ = 'account.move.line'
@@ -37,6 +36,9 @@ class MoveLine(metaclass=PoolMeta):
             depends=['payment_kind', 'second_currency_digits',
                 'currency_digits']), 'get_payment_amount',
         searcher='search_payment_amount')
+    payment_currency = fields.Function(fields.Many2One(
+            'currency.currency', "Payment Currency"),
+        'get_payment_currency')
     payments = fields.One2Many('account.payment', 'line', 'Payments',
         readonly=True,
         states={
@@ -45,8 +47,7 @@ class MoveLine(metaclass=PoolMeta):
         depends=['payment_kind'])
     payment_kind = fields.Function(fields.Selection([
                 (None, ''),
-                ] + KINDS, 'Payment Kind'), 'get_payment_kind',
-        searcher='search_payment_kind')
+                ] + KINDS, 'Payment Kind'), 'get_payment_kind')
     payment_blocked = fields.Boolean('Blocked', readonly=True)
     payment_direct_debit = fields.Boolean("Direct Debit",
         states={
@@ -62,7 +63,8 @@ class MoveLine(metaclass=PoolMeta):
         super(MoveLine, cls).__setup__()
         cls._buttons.update({
                 'pay': {
-                    'invisible': ~Eval('payment_kind').in_(list(dict(KINDS).keys())),
+                    'invisible': ~Eval('payment_kind').in_(
+                        list(dict(KINDS).keys())),
                     'depends': ['payment_kind'],
                     },
                 'payment_block': {
@@ -85,7 +87,8 @@ class MoveLine(metaclass=PoolMeta):
     def get_payment_amount(cls, lines, name):
         amounts = {}
         for line in lines:
-            if line.account.kind not in ('payable', 'receivable'):
+            if (not line.account.type.payable
+                    and not line.account.type.receivable):
                 amounts[line.id] = None
                 continue
             if line.second_currency:
@@ -105,11 +108,13 @@ class MoveLine(metaclass=PoolMeta):
         pool = Pool()
         Payment = pool.get('account.payment')
         Account = pool.get('account.account')
+        AccountType = pool.get('account.account.type')
         _, operator, value = clause
         Operator = fields.SQL_OPERATORS[operator]
         table = cls.__table__()
         payment = Payment.__table__()
         account = Account.__table__()
+        account_type = AccountType.__table__()
 
         payment_amount = Sum(Coalesce(payment.amount, 0))
         main_amount = Abs(table.credit - table.debit) - payment_amount
@@ -118,22 +123,31 @@ class MoveLine(metaclass=PoolMeta):
             else_=second_amount)
         value = cls.payment_amount.sql_format(value)
 
-        query = table.join(payment, type_='LEFT',
-            condition=(table.id == payment.line) & (payment.state != 'failed')
-            ).join(account, condition=table.account == account.id
-                ).select(table.id,
-                    where=account.kind.in_(['payable', 'receivable']),
-                    group_by=(table.id, account.kind, table.second_currency),
-                    having=Operator(amount, value)
-                    )
+        query = (table
+            .join(payment, type_='LEFT',
+            condition=(table.id == payment.line) & (payment.state != 'failed'))
+            .join(account, condition=table.account == account.id)
+            .join(account_type, condition=account.type == account_type.id)
+            .select(table.id,
+                where=(account_type.payable | account_type.receivable),
+                group_by=(table.id, table.second_currency),
+                having=Operator(amount, value)
+                ))
         return [('id', 'in', query)]
 
-    def get_payment_kind(self, name):
-        return self.account.kind if self.account.kind in dict(KINDS) else None
+    def get_payment_currency(self, name):
+        if self.second_currency:
+            return self.second_currency.id
+        elif self.currency:
+            return self.currency.id
 
-    @classmethod
-    def search_payment_kind(cls, name, clause):
-        return [('account.kind',) + tuple(clause[1:])]
+    def get_payment_kind(self, name):
+        if (self.account.type.receivable
+                and (self.debit > 0 or self.credit < 0)):
+            return 'receivable'
+        elif (self.account.type.payable
+                and (self.credit > 0 or self.debit < 0)):
+            return 'payable'
 
     @classmethod
     def default_payment_blocked(cls):
@@ -175,6 +189,15 @@ class MoveLine(metaclass=PoolMeta):
                 })
 
 
+class PayLineStart(ModelView):
+    "Pay Line"
+    __name__ = 'account.move.line.pay.start'
+    date = fields.Date(
+        "Date",
+        help="When the payments are scheduled to happen.\n"
+        "Leave empty to use the lines' maturity dates.")
+
+
 class PayLineAskJournal(ModelView):
     'Pay Line'
     __name__ = 'account.move.line.pay.ask_journal'
@@ -194,20 +217,70 @@ class PayLineAskJournal(ModelView):
 class PayLine(Wizard):
     'Pay Line'
     __name__ = 'account.move.line.pay'
-    start = StateTransition()
+    start = StateView(
+        'account.move.line.pay.start',
+        'account_payment.move_line_pay_start_view_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Pay", 'next_', 'tryton-ok', default=True),
+            ])
+    next_ = StateTransition()
     ask_journal = StateView('account.move.line.pay.ask_journal',
         'account_payment.move_line_pay_ask_journal_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Pay', 'start', 'tryton-ok', default=True),
+            Button('Pay', 'next_', 'tryton-ok', default=True),
             ])
     pay = StateAction('account_payment.act_payment_form')
 
-    @classmethod
-    def __setup__(cls):
-        super(PayLine, cls).__setup__()
-        cls._error_messages.update({
-                'blocked': 'The Line "%(line)s" is blocked.',
-                })
+    def default_start(self, fields):
+        pool = Pool()
+        Line = pool.get('account.move.line')
+        Warning = pool.get('res.user.warning')
+
+        reverse = {'receivable': 'payable', 'payable': 'receivable'}
+        companies = {}
+        lines = self.records
+        for line in lines:
+            types = companies.setdefault(line.move.company, {
+                    kind: {
+                        'parties': set(),
+                        'lines': list(),
+                        }
+                    for kind in reverse.keys()})
+            for kind in types:
+                if getattr(line.account.type, kind):
+                    types[kind]['parties'].add(line.party)
+                    types[kind]['lines'].append(line)
+
+        for company, types in companies.items():
+            for kind in types:
+                parties = types[kind]['parties']
+                others = Line.search([
+                        ('move.company', '=', company.id),
+                        ('account.type.' + reverse[kind], '=', True),
+                        ('party', 'in', [p.id for p in parties]),
+                        ('reconciliation', '=', None),
+                        ('payment_amount', '!=', 0),
+                        ('move_state', '=', 'posted'),
+                        ])
+                for party in parties:
+                    party_lines = [l for l in others if l.party == party]
+                    if not party_lines:
+                        continue
+                    lines = [l for l in types[kind]['lines']
+                        if l.party == party]
+                    warning_name = '%s:%s:%s' % (
+                        reverse[kind], party,
+                        hashlib.md5(str(lines).encode('utf-8')).hexdigest())
+                    if Warning.check(warning_name):
+                        names = ', '.join(l.rec_name for l in lines[:5])
+                        if len(lines) > 5:
+                            names += '...'
+                        raise GroupWarning(warning_name,
+                            gettext('account_payment.msg_pay_line_group',
+                                names=names,
+                                party=party.rec_name,
+                                line=party_lines[0].rec_name))
+        return {}
 
     def _get_journals(self):
         journals = {}
@@ -230,10 +303,7 @@ class PayLine(Wizard):
             return (company, currency)
 
     def _missing_journal(self):
-        pool = Pool()
-        Line = pool.get('account.move.line')
-
-        lines = Line.browse(Transaction().context['active_ids'])
+        lines = self.records
         journals = self._get_journals()
 
         for line in lines:
@@ -241,7 +311,7 @@ class PayLine(Wizard):
             if key not in journals:
                 return key
 
-    def transition_start(self):
+    def transition_next_(self):
         if self._missing_journal():
             return 'ask_journal'
         else:
@@ -273,8 +343,7 @@ class PayLine(Wizard):
         else:
             kind = 'payable'
         journal = journals[self._get_journal_key(line)]
-
-        return Payment(
+        payment = Payment(
             company=line.move.company,
             journal=journal,
             party=line.party,
@@ -282,22 +351,28 @@ class PayLine(Wizard):
             amount=line.payment_amount,
             line=line,
             )
+        date = self.start.date or line.maturity_date
+        # Use default value when empty
+        if date:
+            payment.date = date
+        return payment
 
     def do_pay(self, action):
         pool = Pool()
-        Line = pool.get('account.move.line')
         Payment = pool.get('account.payment')
+        Warning = pool.get('res.user.warning')
 
-        lines = Line.browse(Transaction().context['active_ids'])
+        lines = self.records
         journals = self._get_journals()
 
         payments = []
         for line in lines:
             if line.payment_blocked:
-                self.raise_user_warning('blocked:%s' % line,
-                    'blocked', {
-                        'line': line.rec_name,
-                        })
+                warning_name = 'blocked:%s' % line
+                if Warning.check(warning_name):
+                    raise BlockedWarning(warning_name,
+                        gettext('account_payment.msg_pay_line_blocked',
+                            line=line.rec_name))
             payments.append(self.get_payment(line, journals))
         Payment.save(payments)
         return action, {
@@ -312,7 +387,9 @@ class Configuration(metaclass=PoolMeta):
             domain=[
                 ('company', 'in',
                     [Eval('context', {}).get('company', -1), None]),
-                ('code', '=', 'account.payment.group'),
+                ('sequence_type', '=',
+                    Id('account_payment',
+                        'sequence_type_account_payment_group')),
                 ]))
 
     @classmethod
@@ -328,14 +405,14 @@ class ConfigurationPaymentGroupSequence(ModelSQL, CompanyValueMixin):
         'ir.sequence', "Payment Group Sequence", required=True,
         domain=[
             ('company', 'in', [Eval('company', -1), None]),
-            ('code', '=', 'account.payment.group'),
+            ('sequence_type', '=',
+                Id('account_payment', 'sequence_type_account_payment_group')),
             ],
         depends=['company'])
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        exist = TableHandler.table_exist(cls._table)
+        exist = backend.TableHandler.table_exist(cls._table)
 
         super(ConfigurationPaymentGroupSequence, cls).__register__(module_name)
 
@@ -368,8 +445,9 @@ class Invoice(metaclass=PoolMeta):
     payment_direct_debit = fields.Boolean("Direct Debit",
         states={
             'invisible': Eval('type') != 'in',
+            'readonly': Eval('state') != 'draft',
             },
-        depends=['type'],
+        depends=['type', 'state'],
         help="Check if the invoice is paid by direct debit.")
 
     @classmethod
@@ -392,25 +470,27 @@ class Invoice(metaclass=PoolMeta):
         pool = Pool()
         Currency = pool.get('currency.currency')
         Date = pool.get('ir.date')
+        context = Transaction().context
 
         today = Date.today()
 
         amounts = super(Invoice, cls).get_amount_to_pay(invoices, name)
 
-        for invoice in invoices:
-            for line in invoice.lines_to_pay:
-                if line.reconciliation:
-                    continue
-                if (name == 'amount_to_pay_today'
-                        and line.maturity_date
-                        and line.maturity_date > today):
-                    continue
-                payment_amount = Decimal(0)
-                for payment in line.payments:
-                    if payment.state != 'failed':
-                        with Transaction().set_context(date=payment.date):
-                            payment_amount += Currency.compute(
-                                payment.currency, payment.amount,
-                                invoice.currency)
-                amounts[invoice.id] -= payment_amount
+        if context.get('with_payment', True):
+            for invoice in invoices:
+                for line in invoice.lines_to_pay:
+                    if line.reconciliation:
+                        continue
+                    if (name == 'amount_to_pay_today'
+                            and line.maturity_date
+                            and line.maturity_date > today):
+                        continue
+                    payment_amount = Decimal(0)
+                    for payment in line.payments:
+                        if payment.state != 'failed':
+                            with Transaction().set_context(date=payment.date):
+                                payment_amount += Currency.compute(
+                                    payment.currency, payment.amount,
+                                    invoice.currency)
+                    amounts[invoice.id] -= payment_amount
         return amounts
